from clipped.utils.enums import PEnum


class PatchStrategy(str, PEnum):
    REPLACE = "replace"
    ISNULL = "isnull"
    POST_MERGE = "post_merge"
    PRE_MERGE = "pre_merge"

    @classmethod
    def is_replace(cls, value):
        return value == cls.REPLACE

    @classmethod
    def is_null(cls, value):
        return value == cls.ISNULL

    @classmethod
    def is_post_merge(cls, value):
        return value == cls.POST_MERGE

    @classmethod
    def is_pre_merge(cls, value):
        return value == cls.PRE_MERGE



from typing import Dict, List, Optional, Tuple

from clipped.utils.humanize import humanize_attrs


def deep_update(config: Dict, override_config: Dict):
    for k, v in override_config.items():
        if isinstance(v, dict):
            k_config = config.get(k, {})
            if isinstance(k_config, dict):
                v_config = deep_update(k_config, v)
                config[k] = v_config
            else:
                config[k] = v
        else:
            config[k] = override_config[k]
    return config


def flatten_keys(
    objects: List[Dict], columns: List[str], columns_prefix: Optional[Dict] = None
) -> Tuple[List[Dict], Dict]:
    # Extend run with params_keys
    keys = set([])
    columns_prefix = columns_prefix or {}
    prefixed_columns = {}

    def process_objects():
        results = {}
        for k, v in col_values.items():
            results["{}.{}".format(column_prefix, k)] = v
            prefixed_columns[k] = "{}.{}".format(column_prefix, k)
        return results

    for col in columns:
        for obj in objects:
            col_values = obj.pop(col, {}) or {}
            if col in columns_prefix:
                column_prefix = columns_prefix[col]
                col_values = process_objects()
            col_keys = col_values.keys()
            keys |= set(col_keys)
            obj.update(col_values)

    # Check that all obj have all metrics
    # TODO: optimize this process
    for obj in objects:
        obj_keys = set(obj.keys())
        for key in keys:
            if key not in obj_keys:
                obj[key] = None

    return objects, prefixed_columns


def list_dicts_to_tabulate(
    list_dicts,
    exclude_attrs=None,
    include_attrs=None,
    humanize_values=True,
    upper_keys: bool = True,
):
    exclude_attrs = exclude_attrs or {}
    results = []
    if include_attrs:  # If include_attrs disable exclude_attrs
        exclude_attrs = {}
    for d_value in list_dicts:
        r_value = {}
        for k, v in d_value.items():
            if k in exclude_attrs:
                continue
            if include_attrs and k not in include_attrs:
                continue

            if humanize_values:
                v = humanize_attrs(k, v)

            if upper_keys:
                k = k.upper()
            r_value[k] = v
        results.append(r_value)
    return results


def list_dicts_to_csv(
    list_dicts,
    exclude_attrs=None,
    include_attrs=None,
):
    exclude_attrs = exclude_attrs or {}
    results = []
    if include_attrs:  # If include_attrs disable exclude_attrs
        exclude_attrs = {}
    for d_value in list_dicts:
        result = {}
        for k, v in d_value.items():
            if k in exclude_attrs:
                continue
            if include_attrs and k not in include_attrs:
                continue
            result[k] = v
        results.append(result)

    return results


def dict_to_tabulate(
    d_value,
    exclude_attrs=None,
    humanize_values=True,
    rounding: int = 2,
    timesince: bool = True,
):
    exclude_attrs = exclude_attrs or {}
    results = {}
    for k, v in d_value.items():
        if k in exclude_attrs:
            continue

        if humanize_values:
            v = humanize_attrs(k, v, rounding=rounding, timesince=timesince)

        results[k.upper()] = v

    return results



from typing import Dict, Optional


def get_query_params(
    limit: Optional[int] = None,
    offset: Optional[int] = None,
    query: Optional[str] = None,
    sort: Optional[str] = None,
) -> Dict:
    params = {}
    if limit:
        params["limit"] = limit
    if offset:
        params["offset"] = offset
    if query:
        params["query"] = query
    if sort:
        params["sort"] = sort

    return params


def get_logs_params(
    last_time: Optional[str] = None, last_file: Optional[str] = None
) -> Dict:
    params = {}
    if last_file:
        params["last_file"] = last_file
    if last_time:
        params["last_time"] = last_time

    return params



from typing import TYPE_CHECKING, Any, Dict

from pydantic import AnyUrl

if TYPE_CHECKING:
    from pydantic.config import BaseConfig
    from pydantic.fields import ModelField


class Uri(AnyUrl):
    __slots__ = ()

    @classmethod
    def validate(
        cls, value: Any, field: "ModelField", config: "BaseConfig"
    ) -> "AnyUrl":
        if isinstance(value, Dict):
            _value = value.get("user")
            if not _value:
                raise ValueError("Received a wrong bucket definition: %s", value)
            password = value.get("password")
            if password:
                _value = "{}@{}".format(_value, password)
            host = value.get("host")
            if not host:
                raise ValueError("Received a wrong bucket definition: %s", value)
            _value = "{}/{}".format(_value, host)
            value = _value
        return super(Uri, cls).validate(value=value, field=field, config=config)

    def to_param(self):
        return str(self)

    @property
    def host_port(self):
        value = self.host
        if self.port:
            value = "{}:{}".format(value, self.port)
        if self.scheme:
            value = "{}://{}".format(self.scheme, value)
        return value


V1UriType = Uri



import datetime

from typing import Any, Dict

from pydantic import (
    ByteSize,
    ConstrainedBytes,
    ConstrainedDate,
    ConstrainedDecimal,
    ConstrainedFloat,
    ConstrainedFrozenSet,
    ConstrainedInt,
    ConstrainedList,
    ConstrainedSet,
    ConstrainedStr,
    FiniteFloat,
    FutureDate,
    Json,
    NegativeFloat,
    NegativeInt,
    NoneBytes,
    NoneStr,
    NoneStrBytes,
    NonNegativeFloat,
    NonNegativeInt,
    NonPositiveFloat,
    NonPositiveInt,
    PastDate,
    PaymentCardNumber,
    PositiveFloat,
    PositiveInt,
    PyObject,
    SecretBytes,
    SecretField,
    SecretStr,
    StrBytes,
    StrictBool,
    StrictBytes,
    StrictFloat,
    StrictInt,
    StrictStr,
)

from clipped.types.docker_image import ImageStr
from clipped.types.email import EmailStr
from clipped.types.gcs import GcsPath
from clipped.types.lists import ListStr
from clipped.types.s3 import S3Path
from clipped.types.strings import GenericStr
from clipped.types.uri import Uri
from clipped.types.uuids import UUIDStr
from clipped.types.wasb import WasbPath

URI = "uri"
AUTH = "auth"
LIST = "list"
GCS = "gcs"
S3 = "s3"
WASB = "wasb"
IMAGE = "image"
PATH = "path"
METRIC = "metric"
METADATA = "metadata"
DATE = "date"
DATETIME = "datetime"
UUID = "uuid"
EMAIL = "email"

MAPPING = {
    URI: Uri,
    IMAGE: ImageStr,
    GCS: GcsPath,
    S3: S3Path,
    WASB: WasbPath,
    PATH: str,
    METRIC: float,
    METADATA: Dict,
    UUID: UUIDStr,
    EMAIL: EmailStr,
    "any": Any,
    "str": GenericStr,
    "date": datetime.date,
    "datetime": datetime.date,
    "timedelta": datetime.timedelta,
}


FORWARDING = {
    "GenericStr": GenericStr,
    "Uri": Uri,
    "ImageStr": ImageStr,
    "GcsPath": GcsPath,
    "S3Path": S3Path,
    "WasbPath": WasbPath,
    "UUIDStr": UUIDStr,
    "EmailStr": EmailStr,
    "ListStr": ListStr,
    "NoneStr": NoneStr,
    "NoneBytes": NoneBytes,
    "StrBytes": StrBytes,
    "NoneStrBytes": NoneStrBytes,
    "StrictStr": StrictStr,
    "ConstrainedBytes": ConstrainedBytes,
    "ConstrainedList": ConstrainedList,
    "ConstrainedSet": ConstrainedSet,
    "ConstrainedFrozenSet": ConstrainedFrozenSet,
    "ConstrainedStr": ConstrainedStr,
    "PyObject": PyObject,
    "ConstrainedInt": ConstrainedInt,
    "PositiveInt": PositiveInt,
    "NegativeInt": NegativeInt,
    "NonNegativeInt": NonNegativeInt,
    "NonPositiveInt": NonPositiveInt,
    "ConstrainedFloat": ConstrainedFloat,
    "PositiveFloat": PositiveFloat,
    "NegativeFloat": NegativeFloat,
    "NonNegativeFloat": NonNegativeFloat,
    "NonPositiveFloat": NonPositiveFloat,
    "FiniteFloat": FiniteFloat,
    "ConstrainedDecimal": ConstrainedDecimal,
    "Json": Json,
    "SecretField": SecretField,
    "SecretStr": SecretStr,
    "SecretBytes": SecretBytes,
    "StrictBool": StrictBool,
    "StrictBytes": StrictBytes,
    "StrictInt": StrictInt,
    "StrictFloat": StrictFloat,
    "PaymentCardNumber": PaymentCardNumber,
    "ByteSize": ByteSize,
    "PastDate": PastDate,
    "FutureDate": FutureDate,
    "ConstrainedDate": ConstrainedDate,
}

NON_LOADABLE = {
    "any",
    "str",
    "bool",
    "path",
    "date",
    "datetime",
    "timedelta",
    "uuid",
    "email",
    "Any",
    Any,
    str,
    bool,
    IMAGE,
    ImageStr,
    URI,
    Uri,
    GCS,
    GcsPath,
    S3,
    S3Path,
    WASB,
    WasbPath,
    EmailStr,
    UUIDStr,
    ListStr,
    GenericStr,
    datetime.date,
    datetime.datetime,
    datetime.timedelta,
}


COMPATIBLE_TYPES = [
    ["str", PATH, S3, GCS, WASB],
    [float, METRIC],
]


def are_compatible(type1: str, type2: str) -> bool:
    if type1 == type2:
        return True

    if type1 == "Any" and type2 == "Any":
        return True
    # Compatible types
    for compatible_type in COMPATIBLE_TYPES:
        if type1 in compatible_type and type2 in compatible_type:
            return True

    return False



from typing import TYPE_CHECKING
from uuid import UUID

from pydantic import StrictStr

if TYPE_CHECKING:
    from pydantic.typing import CallableGenerator


class UUIDStr(StrictStr):
    @classmethod
    def __get_validators__(cls) -> "CallableGenerator":
        yield cls.validate

    @classmethod
    def validate(cls, value, **kwargs):
        if isinstance(value, str):
            return UUID(value).hex
        if isinstance(value, UUID):
            return value.hex
        if not value:
            return value

        field = kwargs.get("field")
        raise TypeError(
            f"Field `{field.name}` value must be a valid UUID, received `{value}` instead."
        )



from typing import TYPE_CHECKING

from clipped.utils.lists import to_list

if TYPE_CHECKING:
    from pydantic.typing import CallableGenerator


class ListStr(list):
    @classmethod
    def __get_validators__(cls) -> "CallableGenerator":
        yield cls.validate

    @classmethod
    def validate(cls, value, **kwargs):
        if isinstance(value, list):
            return value
        if isinstance(value, str):
            return to_list(value, check_none=True, check_str=True)
        if not value:
            return value

        field = kwargs.get("field")
        raise TypeError(
            f"Field `{field.name}` value must be a valid List, received `{value}` instead."
        )



from typing import TYPE_CHECKING, Any, Dict

from pydantic import AnyUrl

if TYPE_CHECKING:
    from pydantic.config import BaseConfig
    from pydantic.fields import ModelField


class S3Path(AnyUrl):
    allowed_schemes = {"s3"}

    __slots__ = ()

    @classmethod
    def validate(
        cls, value: Any, field: "ModelField", config: "BaseConfig"
    ) -> "AnyUrl":
        if isinstance(value, Dict):
            _value = value.get("bucket")
            if not _value:
                raise ValueError("Received a wrong bucket definition: %s", value)
            if "://" not in _value:
                _value = "s3://{}".format(_value)
            key = value.get("key")
            if key:
                _value = "{}/{}".format(_value, key)
            value = _value
        return super(S3Path, cls).validate(value=value, field=field, config=config)

    def to_param(self):
        return str(self)

    @property
    def structured(self):
        return dict(bucket=self.host, key=self.path.strip("/"))
